/* This is a function which takes the data written under a CIP loop_
 * directive and places it in a dynamically allocated data structure.
 *
 * This function is supposed to be used as follows:  
 *
 * The calling function hits a 'loop_' directive and passes control
 * to this function.  The first lines of the CIF file below the
 * loop_ directive are should be cif data names.  Function returns
 * NULL if there was a memory allocation failure or if the cif 
 * category string is not matched.
 */
#define _POSIX_SOURCE

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

/* local headers */
#include "ProcessCIFLoop.h"

struct cif_loop *ProcessCIFLoop( FILE *cif, char *category )
{
   size_t len;
   int in_value = 0;
   char cif_line_read[CIF_LINE_LEN]; 

   char *ptr;                          /* general pointer to char for various op's */

/* make it a pointer to a structure so that the pointer value can be
 * returned to the caller.
 */
   struct cif_loop *loop;

   loop = malloc( sizeof(struct cif_loop) );
   if( !loop ) return NULL;

/* initialize the loop structure */
   loop->n_fields = 0;
   loop->n_rows = 0;

   loop->head_anchor.next = malloc( sizeof(struct cif_loop_header) );
   if( !loop->head_anchor.next) MemoryFailure( __FILE__, __LINE__ );

   loop->header = loop->head_anchor.next;
   loop->header->next = NULL;

   loop->row_anchor.next = malloc( sizeof(struct cif_loop_row) );
   if( !loop->row_anchor.next) MemoryFailure( __FILE__, __LINE__ );

   loop->data = loop->row_anchor.next;
   loop->data->next = NULL;

   while( NULL != fgets(cif_line_read, CIF_LINE_LEN, cif) ) {
      /* squash newline and squash everything after comment (#) symbol
       * this is a potential bug:  if a '#' is in text variable, it
       * will blow away everything on that line after the '#',
       * oh well, worry about that later.
       */
       if( NULL != (ptr = strrchr(cif_line_read, '\n')) ) *ptr = '\0';
       if( NULL != (ptr = strrchr(cif_line_read, '#')) ) *ptr = '\0';

       ptr = cif_line_read;

       while( *ptr ) { 
           if( isspace(*ptr) ) {  /* skip the leading white space */
               ptr++;
               continue;
           }
      
           if( UNDER_SCORE == *ptr ) {         /* Then we are in a cif data name */
               char *tmp_ptr = ptr;
               if( 0 != strncmp(ptr, category, strlen(category)) ) {
                 free( loop );
                 return NULL;
               }
               loop->n_fields++;
               while( *ptr ) {
                 if( isspace( *ptr ) ) break;
                 *ptr = tolower( *ptr );
                 ptr++;
               }
               len = ptr - tmp_ptr;     /* gives the length of the CIF data name */
               loop->header->field_name = malloc( len + 1 );
               if( !loop->header->field_name ) MemoryFailure( __FILE__, __LINE__ );
               strncpy( loop->header->field_name, tmp_ptr, len );
               loop->header->field_name[len] = '\0';
               loop->header->next = malloc( sizeof(struct cif_loop_header) );
               if( !loop->header->next ) MemoryFailure( __FILE__, __LINE__ );
               loop->header = loop->header->next;
               loop->header->next = NULL;

           } /* end of the cif data name processing */

/* fill the loop structure with data, assume that one row in the
 * loop is on a single line in the CIF file.  This is how most 
 * CIF files are written, but I don't think it is not required by
 * the CIF standard to be as as such, and so could be a source of
 * bugs in future.
 */
           if( isprint(*ptr) ) {
               int j = 0;
               loop->n_rows++;
               len = strlen( ptr );
               strncpy( loop->data->data_line, ptr, len );
               loop->data->data_line[len] = '\0';
               ptr = loop->data->data_line;

               loop->data->field_ptr = malloc( loop->n_fields * sizeof(char **) );
               if( !loop->data->field_ptr )
                   MemoryFailure( __FILE__, __LINE__ );

               while( *ptr ) {
                 if( !isspace(*ptr) && !in_value ) {
                     loop->data->field_ptr[j] = ptr;
                     in_value = 1;
                     j++;
                 }
                 else if( isspace(*ptr) ) {
                     *ptr = '\0';
                     in_value = 0;
                 }
                 else if( in_value ) {
                     ptr++;
                     continue;
                 }
                 ptr++;      
               }
               in_value = 0;
               loop->data->next = malloc( sizeof(struct cif_loop_row) );
               if( !loop->data->next )
                   MemoryFailure( __FILE__, __LINE__ );
               loop->data = loop->data->next;
               loop->data->next = NULL;
           }
           ptr++;
       } /* end the processing of cif_line_read */

   }
   return loop; 
}

/* we definately need a way to free this loop structure cleanly. So here
 * is a function to do that.   Get to this later.
 */

void FreeCIFLoop( struct cif_loop *loop ) {
    return;
}

void MemoryFailure( char *file_name, int line_number ) {
     fprintf( stderr, "Memory Allocation Error: %s:%d. Exiting.\n", 
              file_name, line_number );
     exit( EXIT_FAILURE );
}
